package com.sun.corba.se.impl.orbutil;

import com.sun.corba.se.impl.logging.ORBUtilSystemException;
import com.sun.corba.se.spi.orb.ORB;

public class CacheTable
{
    private boolean noReverseMap;
    static final int INITIAL_SIZE = 16;
    static final int MAX_SIZE = 1073741824;
    int size;
    int entryCount;
    private CacheTable.Entry[] map;
    private CacheTable.Entry[] rmap;
    private ORB orb;
    private ORBUtilSystemException wrapper;
    
    private CacheTable() {
    }
    
    public CacheTable(ORB orb, boolean noReverseMap) {
        this.orb = orb;
        this.wrapper = ORBUtilSystemException.get(orb, "rpc.encoding");
        this.noReverseMap = noReverseMap;
        this.size = 16;
        this.entryCount = 0;
        this.initTables();
    }
    
    private void initTables() {
        this.map = new CacheTable.Entry[this.size];
        this.rmap = (this.noReverseMap ? null : new CacheTable.Entry[this.size]);
    }
    
    private void grow() {
        if (this.size == 1073741824) {
            return;
        }
        CacheTable.Entry[] map = this.map;
        int size = this.size;
        this.size <<= 1;
        this.initTables();
        for (CacheTable.Entry next : map) {
            while (next != null) {
                this.put_table(next.key, next.val);
                next = next.next;
            }
        }
        return;
    }
    
    private int moduloTableSize(int n) {
        n += ~(n << 9);
        n ^= n >>> 14;
        n += n << 4;
        n ^= n >>> 10;
        return n & this.size - 1;
    }
    
    private int hash(Object o) {
        return this.moduloTableSize(System.identityHashCode(o));
    }
    
    private int hash(int n) {
        return this.moduloTableSize(n);
    }
    
    public final void put(Object o, int n) {
        if (this.put_table(o, n)) {
            ++this.entryCount;
            if (this.entryCount > this.size * 3 / 4) {
                this.grow();
            }
        }
    }
    
    private boolean put_table(Object o, int n) {
        int hash = this.hash(o);
        CacheTable.Entry next = this.map[hash];
        while (next != null) {
            if (next.key == o) {
                if (next.val != n) {
                    throw this.wrapper.duplicateIndirectionOffset();
                }
                return false;
            }
            else {
                next = next.next;
            }
        }
        CacheTable.Entry o2 = new CacheTable.Entry(o, n);
        o2.next = this.map[hash];
        this.map[hash] = o2;
        if (!this.noReverseMap) {
            int hash2 = this.hash(n);
            o2.rnext = this.rmap[hash2];
            this.rmap[hash2] = o2;
        }
        return true;
    }
    
    public final boolean containsKey(Object o) {
        return this.getVal(o) != -1;
    }
    
    public final int getVal(Object o) {
        CacheTable.Entry next = this.map[this.hash(o)];
        while (next != null) {
            if (next.key == o) {
                return next.val;
            }
            next = next.next;
        }
        return -1;
    }
    
    public final boolean containsVal(int n) {
        return this.getKey(n) != null;
    }
    
    public final boolean containsOrderedVal(int n) {
        return this.containsVal(n);
    }
    
    public final Object getKey(int n) {
        CacheTable.Entry rnext = this.rmap[this.hash(n)];
        while (rnext != null) {
            if (rnext.val == n) {
                return rnext.key;
            }
            rnext = rnext.rnext;
        }
        return null;
    }
    
    public void done() {
        this.map = null;
        this.rmap = null;
    }
    
    class Entry
    {
        Object key;
        int val;
        CacheTable.Entry next;
        CacheTable.Entry rnext;
        
        public Entry(Object key, int val) {
            this.key = key;
            this.val = val;
            this.next = null;
            this.rnext = null;
        }
    }
}
